# -*- coding: utf-8 -*-
"""Provides base document support."""
import collections
import logging
from bson.json_util import dumps, loads
from tavi.errors import Errors
from tavi.base.fields import BaseField

logger = logging.getLogger(__name__)


def get_field_attr(cls, field):
    """Custom function for retrieving a tavi.field attribute. Handles nested
    attributes for embedded documents as well as embedded lists of documents.

    """
    value = getattr(cls, field)
    if isinstance(value, BaseDocument):
        value = value.field_values
    elif (isinstance(value, collections.MutableSequence) and
            len(value) and isinstance(value[0], BaseDocument)):
        value = [v.field_values for v in value]
    if value == []:
        value = None
    return value


def set_field_attr(cls, field, value):
    """Custom function for setting a tavi.field attribute."""
    field_descriptor = cls._field_descriptors[field]

    if value is not None and hasattr(field_descriptor, '_type'):
        if not isinstance(value, collections.MutableSequence):
            raise ValueError('ListField value must be a sequence.')

        for item in value:
            getattr(cls, field).append(field_descriptor._type(**item))
        return

    if None == value:
        value = field_descriptor.default

    if isinstance(value, dict):
        value = field_descriptor.doc_class(**value)

    setattr(cls, field, value)


class BaseDocumentMetaClass(type):
    """MetaClass for BaseDocuments. Handles initializing the list of fields for
    the BaseDocument.

    """
    def __init__(cls, name, bases, attrs):
        super(BaseDocumentMetaClass, cls).__init__(name, bases, attrs)

        sorted_fields = sorted(
            [field for field in attrs.iteritems()
                if isinstance(field[1], BaseField) and field[1].persist],
            key=lambda i: i[1].creation_order
        )

        cls._field_descriptors = collections.OrderedDict(sorted_fields)


class BaseDocument(object):
    """Base class for Mongo Documents. Provides basic field support."""
    __metaclass__ = BaseDocumentMetaClass

    def __init__(self, **kwargs):
        self._errors = Errors()
        for field in self.fields:
            set_field_attr(self, field, kwargs.get(field))
        for k, v in kwargs.iteritems():
            if k not in self.fields:
                msg = "Ignoring unknown field for %s: %s = '%s'"
                logger.debug(msg, self.__class__.__name__, repr(k), repr(v))
        self.changed_fields = set()

    @property
    def fields(self):
        """Returns the list of fields for the Document."""
        return self._field_descriptors.keys()

    @property
    def field_values(self):
        """Returns a dictionary containing all fields and their values."""
        return {field: get_field_attr(self, field) for field in self.fields}

    @property
    def mongo_field_values(self):
        """Same as field_values except uses the Mongo field names. This way
        the document can have a field name that is different from the
        field name persisted in Mongo.

        """
        return {
            v.name: get_field_attr(self, k)
            for k, v in self._field_descriptors.items()
        }

    @property
    def errors(self):
        """Returns a tavi.Errors object that contains any errors for the
        Document.

        """
        return self._errors

    @property
    def valid(self):
        """Indicates if all the fields in the Document are valid."""
        self.__validate__()
        return 0 == self.errors.count

    def to_json(self, fields=None):
        """Convert Document model object to JSON. Optionally, specify which
        fields should be serialized.

        """
        include_bson_id = True

        if fields:
            if "bson_id" not in fields:
                include_bson_id = False
            else:
                fields.remove("bson_id")

            field_map = {field: self.field_values[field] for field in fields}
        else:
            field_map = self.field_values

        if include_bson_id:
            field_map["id"] = self.bson_id

        return dumps(field_map)

    @classmethod
    def from_json(cls, json_str):
        """Deserialize a JSON string into a Document model object."""
        attrs = loads(json_str)
        return cls(**attrs)

    def __validate__(self):
        """Override for model level validations. This method will be called by
        the #valid property.

        """
        pass



# -*- coding: utf-8 -*-
"""Provides support for dealing with Mongo Documents."""
from bson.objectid import ObjectId
from tavi import Connection
from tavi.base.documents import BaseDocument, BaseDocumentMetaClass
from tavi.commands import Insert, Update
from tavi.errors import TaviConnectionError
from tavi.utils.timer import Timer
import inflection
import logging
import pymongo
import re

logger = logging.getLogger(__name__)


class DocumentMetaClass(BaseDocumentMetaClass):
    """MetaClass for Documents. Sets up the database connection, infers the
    collection name by pluralizing and underscoring the class name, and sets
    the collection fo rthe Document.

    """

    def __init__(cls, name, bases, attrs):
        super(DocumentMetaClass, cls).__init__(name, bases, attrs)
        cls._collection_name = inflection.underscore(
            inflection.pluralize(name))

    @property
    def collection(cls):
        """Returns a handle to the Document collection."""
        if not Connection.database:
            raise TaviConnectionError(
                "Cannot connect to MongoDB. Did you call "
                "'tavi.connection.Connection.setup'?")
        return Connection.database[cls._collection_name]

    @property
    def collection_name(cls):
        """Returns the name of the Document collection."""
        return cls._collection_name


class Document(BaseDocument):
    """Represents a Mongo Document. Provides methods for saving and retrieving
    and deleting Documents.

    """
    __metaclass__ = DocumentMetaClass

    __MAX_NAMESPACE_SIZE__ = 127  # bytes
    __UNIQUE_INDEX_SUFFIX__ = "_unique_index"

    def __init__(self, **kwargs):
        self._id = kwargs.pop("_id", None)
        super(Document, self).__init__(**kwargs)
        unique_keys = [
            k for k, v in self._field_descriptors.items() if v.unique]

        if len(unique_keys) > 0:
            key_pairs = [(k, pymongo.ASCENDING) for k in unique_keys]
            max_index_name_length = self.__MAX_NAMESPACE_SIZE__ - \
                len(".$" + self.__class__.collection.full_name
                    + self.__UNIQUE_INDEX_SUFFIX__)

            name = "_".join(unique_keys)[:max_index_name_length] + \
                self.__UNIQUE_INDEX_SUFFIX__

            opts = {"name": name, "unique": True}
            self.__class__.collection.create_index(key_pairs, **opts)

    @property
    def bson_id(self):
        """Returns the BSON Id of the Document."""
        return self._id

    @classmethod
    def count(cls):
        """Returns the total number of documents in the collection."""
        return cls.collection.count()

    def delete(self):
        """Removes the Document from the collection."""
        timer = Timer()
        with timer:
            result = self.__class__.collection.remove({"_id": self._id})

        logger.info(
            "(%ss) %s DELETE %s",
            timer.duration_in_seconds(),
            self.__class__.__name__,
            self._id
        )

        if result.get("err"):
            logger.error(result.get("err"))

    @classmethod
    def find(cls, *args, **kwargs):
        """Returns all Documents in collection that meet criteria. Wraps
        pymongo's *find* method and supports all of the same arguments.

        """
        timer = Timer()
        with timer:
            results = cls.collection.find(*args, **kwargs)

        logger.info(
            "(%ss) %s FIND %s, %s (%s record(s) found)",
            timer.duration_in_seconds(),
            cls.__name__,
            args,
            kwargs,
            results.count()
        )

        return [cls(**result) for result in results]

    @classmethod
    def find_all(cls):
        """Returns all Documents in collection."""
        return cls.find()

    @classmethod
    def find_by_id(cls, id_):
        """Returns the Document that matches *id_* or None if it cannot be
        found.

        """
        return cls.find_one(ObjectId(id_))

    @classmethod
    def find_one(cls, spec_or_id=None, *args, **kwargs):
        """Returns one Document that meets criteria. Wraps pymongo's find_one
        method and supports all of the same arguments.

        """
        timer = Timer()
        with timer:
            result = cls.collection.find_one(spec_or_id, *args, **kwargs)

        found_record, num_found = None, 0

        if result:
            found_record, num_found = cls(**result), 1

        logger.info(
            "(%ss) %s FIND ONE %s, %s, %s (%s record(s) found)",
            timer.duration_in_seconds(),
            cls.__name__,
            spec_or_id,
            args,
            kwargs,
            num_found
        )

        return found_record

    def save(self, w=1, wtimeout=0, j=False):
        """Saves the Document by inserting it into the collection if it does
        not exist or updating it if it does. Returns True if save was
        successful. Ensures that the Document is valid before saving and
        returns False if it was not.

        This function performs an upsert if the model has an ID, but is not in
        the database.

        If the document model has a field named 'created_at', this field's
        value will be set to the current time when the document is inserted.

        Supports the following arguments that are passed to pymongo:

        w: Write concern level. Default is 1
        wtimeout: Timeout in ms for write concern. Default is 0
        j: Journaling option for write concern. Default is False

        See Pymongo docs for more information on these arguments:
        http://api.mongodb.org/python/current/api/pymongo/collection.html

        See the Mongo docs for more information on Write Concern:
        http://docs.mongodb.org/manual/core/write-concern/

        """
        if not self.valid:
            return False

        write_opts = frozenset(["w", "j", "wtimeout"])
        kwargs = {k: v for k, v in locals().iteritems() if k in write_opts}

        operation = Update if self.bson_id else Insert
        operation = operation(self, **kwargs)

        timer = Timer()
        with timer:
            try:
                operation.execute()
            except pymongo.errors.PyMongoError as e:
                operation.reset_fields()

                if isinstance(e, pymongo.errors.DuplicateKeyError):
                    logger.warn(
                        "%s %s failed due to unique index violation (%s)",
                        self.__class__.__name__,
                        operation.name,
                        e.message
                    )
                    f = re.search(r'\$(.+)_unique_index', e.message).group(1)
                    self.errors.add(f, "must be unique")
                    return False
                raise

        self.changed_fields = set()

        logger.info(
            "(%ss) %s %s %s, %s",
            timer.duration_in_seconds(),
            self.__class__.__name__,
            operation.name,
            self.mongo_field_values,
            self._id
        )
        return True


class EmbeddedDocument(BaseDocument):
    """Represents a single EmbeddedDocument. Supports an *owner* attribute that
    indicates the owning Document.

    """
    def __init__(self, **kwargs):
        super(EmbeddedDocument, self).__init__(**kwargs)
        self.owner = None

    def __eq__(self, other):
        return other and self.field_values == other.field_values



# -*- coding: utf-8 -*-
import collections
import datetime
import tavi.documents


class MongoCommand(object):
    def __init__(self, target, **kwargs):
        self.target = target
        self.kwargs = kwargs

    @property
    def name(self):
        raise "Not Implemented"

    def execute(self):
        self._now = datetime.datetime.utcnow()
        if hasattr(self.target, "last_modified_at"):
            self.old_last_modified_at = self.target.last_modified_at

        self._update_field("last_modified_at", self._now)

    def reset_fields(self):
        if hasattr(self.target, "last_modified_at"):
            self._update_field("last_modified_at", self.old_last_modified_at)

    def _update_field(self, name, timestamp):
        for field in self.target.fields:
            value = getattr(self.target, field)
            if name == field:
                setattr(self.target, name, timestamp)
            elif isinstance(value, collections.Iterable):
                for item in value:
                    if isinstance(item, tavi.documents.EmbeddedDocument):
                        setattr(item, name, timestamp)
            elif isinstance(value, tavi.documents.EmbeddedDocument):
                if hasattr(value, name):
                    setattr(value, name, timestamp)


class Insert(MongoCommand):
    @property
    def name(self):
        return "INSERT"

    def execute(self):
        super(Insert, self).execute()
        if hasattr(self.target, "created_at"):
            self.old_created_at = self.target.created_at

        self._update_field("created_at", self._now)

        collection = self.target.__class__.collection
        values = self.target.mongo_field_values
        self.target._id = collection.insert(values, **self.kwargs)

    def reset_fields(self):
        super(Insert, self).reset_fields()
        if hasattr(self.target, "created_at"):
            self._update_field("created_at", self.old_created_at)


class Update(MongoCommand):
    @property
    def name(self):
        return "UPDATE"

    def execute(self):
        super(Update, self).execute()
        self.kwargs["upsert"] = True
        self.target.__class__.collection.update(
            {"_id": self.target._id},
            {"$set": self.target.mongo_field_values},
            **self.kwargs)



# -*- coding: utf-8 -*-
"""Provides base field support."""


class BaseField(object):
    """Base class for Mongo Document fields.

    name -- the name of the document field

    All *BaseFields* support the following default validations:

    required  -- indicates if the field is required; default is *False*
    default   -- default value for the field; *None* if not given
    choices -- validates field value is a member of specified list
    persist   -- boolean indicating if field should be persisted to Mongo;
                 default is True

    """

    _creation_counter = 0

    def __init__(
        self, name,
        required=False, default=None, choices=None, persist=True, unique=False
    ):
        self.name = name
        self.attribute_name = "_%s" % name
        self.required = required
        self.default = default
        self.choices = choices
        self.persist = persist
        self.unique = unique
        self.creation_order = BaseField._creation_counter
        BaseField._creation_counter += 1

    def __get__(self, instance, owner):
        if self.attribute_name not in instance.__dict__ and self.default:
            self.__set__(instance, self.default)
        return getattr(instance, self.attribute_name)

    def __set__(self, instance, value):
        if None == value and self.required and self.default:
            value = self.default
        self.validate(instance, value)
        setattr(instance, self.attribute_name, value)
        if hasattr(instance, "changed_fields"):
            instance.changed_fields.add(self.name)

    def validate(self, instance, value):
        """Validates the field.

        Subclasses should call *super* and implement their own validation.
        Assumes that *obj* has an *errors* attribute that acts like a
        tavi.errors.Errors class.

        """
        instance.errors.clear(self.name)
        if self.required and None == value:
            instance.errors.add(self.name, "is required")

        if self.choices and value not in self.choices:
            instance.errors.add(self.name, "value must be in list")
